/*********************************************************************
 *  SPDX-License-Identifier: MIT									 *
 *  Copyright (C) 2007-2024 by Helge Kortendieck  alan@kasmithee.de  *
 *                                                                   *
 *  This file is part of siimvi.                                     *
 *********************************************************************/
#include "siv.h"

Siv::Siv(QWidget *parent)
	: QWidget(parent)
{
	setWindowFlags (Qt::FramelessWindowHint);
	setStyleSheet("background-color: black");
	showMaximized();								// does not overlap task bar (fluxbox 2019)
//	showFullScreen();								// uses full screen size, overlaps task bar (fluxbox 2019)
	QScreen *screen = window()->windowHandle()->screen();
	qrFrame = screen->geometry();
	iScrX = qrFrame.width();
	iScrY = qrFrame.height();

	setRead();

	bInit = true;

	slArgs = QApplication::arguments();               // get command line args, 0=app name
	slFiles.clear();
	for (int i = 1; i < slArgs.count(); ++i) {
		sD = slArgs.at(i);
		QFileInfo qfi(sD);
		if (qfi.exists() and qfi.isReadable() and qfi.size() > 0) {
			slFiles.append(sD);
		}
	}

	if (slFiles.count() == 0) {
		setStyleSheet("background-color: lightGray");
		QMessageBox::warning(this, tr("siimvi - No valid file !"), 
									tr("Not one existing and readable file was provided with the command line arguments !\n \n"
										"siimvi does not have a real GUI with a file open dialog.\n \n"
										"You must invoke the program with at least one image to show in the command line. "
										"If there are other image files in the directory of the first, you can then cycle "
										"between them with the mouse wheel.\n"
										"Or give a number of image files, space separated, on the command line and you can "
										"cycle between these.\n \n"
										"When an image is shown, right clicking into it will open a menu to access further "
										"commands. Among those are the about dialog and the online help.\n \n"
										"Other command line options are -v or --version to show the version number of the "
										"program or -h or --help to show a short help message. But these will only print "
										"to standard out, the console.\n \n"
										"Aborting...\n"));
		quitDo();
	} else {
		timerHideCurCreate();
		timerAnimNextCreate();
		iImCurListIdx = 0;					// needed until init done for 1st image
		usrImageNew(slFiles.at(0));			// load and show image siimvi was called with
		init();								// create actions, menu, tooltips and scan dir of 1st image for more
	}
}

Siv::~Siv()
{
}

void Siv::setRead()
{
	QSettings settings("siimvi", "siimvi");

	settings.beginGroup("main");
		bFTS = settings.value("showfittoscreen", "true").toBool();		// 0=1:1, 1=fit to screen
		bShowName = settings.value("showname", "true").toBool();		// show image name on image open, 0=no, 1=yes
		bShowSize = settings.value("showsize", "true").toBool();		// show image size on image open, 0=no, 1=yes
		bExif = settings.value("autorotateonexif", "true").toBool();	// use exif info to auto-rotate image on image open, 0=no, 1=yes
		sDisp = settings.value("displaytoshowimage", "0:0").toString();	// display to show image on open, default = 0:0
		sDir = settings.value("savetodir", "/tmp").toString();			// dir to save images to, updated on last used dir
    // alwaysopeninfo & displaytoshowinfo not implemented yet
	settings.endGroup();
}

void Siv::setWrite()
{
	QSettings settings("siimvi", "siimvi");

	settings.beginGroup("main");
		settings.setValue("savetodir", sDir);	// dir to save images to, updated on last used dir
	settings.endGroup();

	settings.sync();
}

void Siv::quitPrep()						// quit clicked clean up
{
	if (bWInfo == true) {
		imInfoClose();
	}
	quitDo();									// only for now...
}

void Siv::quitDo()							// finally quit the program
{
  exit(0);
}


// ****************************************************************
// **  init stuff, setting things up                             **
// ****************************************************************
void Siv::init()                            // set everything up (nearly), to allow a fast start, do after showing first image !
{
  bInit = false;
  bBlockWheel = false;      // block loading further images until next image is loaded
  bFiles = false;           // got file or files? Look for others in dir or cycle through given files with mouse wheel up/down
  bGrab = false;            // current grab status: 0=no, 1=yes
  bQuit = false;
  bWAhw = false;            // abouthelp widget created ?
  bWInfo = false;           // info widget created ?
  bWQuest = false;
  bWSetDlg = false;         // settings dialog created ?

  iWShown = 0;              // popup widget currently shown: 0=none, 1=about-abouthelp, 2=help-abouthelp, 3=error-abouthelp, 4=info, 5=settings dialog

  setMouseTracking(true);

  stt = new tT();                                                       // show  at begin/end of image list
  connect(this, SIGNAL(sigShowListBeginEnd(QPoint, const QString&)), stt, SLOT(tShow(QPoint, const QString&)));
  connect(this, SIGNAL(sigSetReadTt()), stt, SLOT(setRead()));    // re-read settings

  sns = new tT();                                                       // show tooltip with name and/or image size
  connect(this, SIGNAL(sigShowNameSize(int, int , const QString&)), sns, SLOT(tShowNS(int, int , const QString&)));
  connect(this, SIGNAL(sigSetReadTtNS()), sns, SLOT(setRead()));  // re-read settings

  defActions();                           // define menu actions
  menu = new QMenu;                       // mini menu to delete item after right mouse click in table
  menuPop();                              // populate the menu

  if (slFiles.count() == 1) {
    readDir();                            // get all images in dir of image called, only if not list of images were given on start
  }
  if (iImLastGood > -1) {
    showImNameSize();                           // now show info of current image if wished
  }
}

void Siv::defActions()
{
	actInfo = new QAction(QIcon("/usr/share/siimvi/icons/scalable/info.svg"), tr("Show image &information"), this);
	actInfo->setStatusTip(tr("Show information for this image"));
	connect(actInfo, SIGNAL(triggered()), this, SLOT(imInfoShow()));  		// info, i-key, action for menu to show image information

	actInfoClose = new QAction(QIcon("/usr/share/siimvi/icons/scalable/x-red.svg"), tr("&Close image information"), this);
	actInfoClose->setStatusTip(tr("Close image information widget"));
	connect(actInfoClose, SIGNAL(triggered()), this, SLOT(imInfoClose()));  // close info, action for menu to close image information

	actLeft = new QAction(QIcon("/usr/share/siimvi/icons/scalable/rotate-left.svg"), tr("Rotate image 90° &left"), this);
	actLeft->setStatusTip(tr("Rotate image 90° left (counter-clockwise)"));
	connect(actLeft, SIGNAL(triggered()), this, SLOT(menuLeft()));    		// left, l-key, action for menu to rotate image 90° left

	actRight = new QAction(QIcon("/usr/share/siimvi/icons/scalable/rotate-right.svg"), tr("Rotate image 90° &right"), this);
	actRight->setStatusTip(tr("Rotate image 90° right (clockwise)"));
	connect(actRight, SIGNAL(triggered()), this, SLOT(menuRight()));  		// right, r-key, action for menu to rotate image 90° right

	actView0 = new QAction(QIcon("/usr/share/siimvi/icons/scalable/show-original.svg"), tr("Show image f&ull size (1:1) - left mouse button"), this);
	actView0->setStatusTip(tr("Show image in original size, might not fit on screen"));
	connect(actView0, SIGNAL(triggered()), this, SLOT(menuToggleFitToScreen()));	// full view, f-key, action for menu to show image 1:1

	actView1 = new QAction(QIcon("/usr/share/siimvi/icons/scalable/show-fullscreen.svg"), tr("Show image &fit to screen - left mouse button"), this);
	actView1->setStatusTip(tr("Show image maximized to screen, might fill full screen"));
	connect(actView1, SIGNAL(triggered()), this, SLOT(menuToggleFitToScreen()));	// fit to screen, f-key, action for menu to show image fit to screen

	actSaveImg = new QAction(QIcon("/usr/share/siimvi/icons/scalable/save.svg"), tr("Save I&mage"), this);
	actSaveImg->setStatusTip(tr("Save this image"));
	connect(actSaveImg, SIGNAL(triggered()), this, SLOT(saveImg()));  		// save the current image, m-key, action to save the current image

	actDlg = new QAction(QIcon("/usr/share/siimvi/icons/scalable/settings.svg"), tr("Settings &dialog"), this);
	actDlg->setStatusTip(tr("Open settings dialog to set some default values"));
	connect(actDlg, SIGNAL(triggered()), this, SLOT(setDlgShow()));			// settings dialog, d-key, action for menu to open settings dialog

	actHelp = new QAction(QIcon("/usr/share/siimvi/icons/scalable/help.svg"), tr("&Help"), this);
	actHelp->setStatusTip(tr("Open help dialog"));
	connect(actHelp, SIGNAL(triggered()), this, SLOT(help()));				// help, h-key, action for menu to open help dialog

	actAbout = new QAction(QIcon("/usr/share/siimvi/icons/scalable/about.svg"), tr("&About"), this);
	actAbout->setStatusTip(tr("Open about dialog"));
	connect(actAbout, SIGNAL(triggered()), this, SLOT(about()));      		// about, a-key, action for menu to open about dialog

	actQuit = new QAction(QIcon("/usr/share/siimvi/icons/scalable/exit.svg"), tr("&Quit"), this);
	actQuit->setStatusTip(tr("Quit the program"));
	connect(actQuit, SIGNAL(triggered()), this, SLOT(quitPrep()));    		// quit, q-key, action for menu to exit program
}

void Siv::readDir()                         // get all images in dir of image called
{
//  cout << "imShowPrep(): " << slFiles.at( 0 ).toStdString() << endl;                                               // *****  debug  *****
//  sImCald = slFiles.at( 0 );
	QFileInfo fi(sImCur);
	sD = fi.path() + "/";
	QDir dir(sD);								// image directory, look for more images if only one image was passed as command line option
		slFilters << "*.GIF" << "*.JPG" << "*.JPEG" << "*.PNG" << "*.SVG" << "*.TIF" << "*.TIFF" << "*.BMP" << "*.PBM" << "*.PGM" << "*.PPM" << "*.XBM" << "*.XPM"; // << "*.CR2" << "*.PNM" 
		dir.setNameFilters(slFilters);
		dir.setFilter(QDir::Files | QDir::Readable);
		dir.setSorting(QDir::Name | QDir::IgnoreCase);
	slD = dir.entryList();
	slFiles.clear();
	if (slD.count() > 0) {						// found at least 1 image with filter settings
		for (int i = 0; i < slD.count(); ++i) {
			slFiles.append(sD + slD.at(i));
		}
		bD = false;
		for (int i = 0; i < slFiles.count(); ++i) {
			if (slFiles.at(i) == sImCur) {	// image siimvi was called with is among the list or files returned by dir serach
				bD = true;						// found image siimvi was called with in image list
				iImCurListIdx = i;
				break;
			}
		}
		if (bD == false) {						// did not found image siimvi was called with in image list
			slFiles.append(sImCur);			// was called with different image type we support (and set filters ), let's try anyway, add at end
			iImCurListIdx = slFiles.count() - 1;
		}
	} else {
		slFiles.append(sImCur);				// was called with different image type we support (and set filters ), let's try anyway
		iImCurListIdx = 0;						// file siimvi was invoked with is only file in list
	}
}


// ****************************************************************
// **  menu population, pre-handling of menu selection           **
// ****************************************************************
void Siv::menuPop()							// populate menu
{
	menu->clear();
	if (bWInfo == true) {
		menu->addAction(actInfoClose);
	} else {
		menu->addAction(actInfo);
	}
	menu->addSeparator();
	menu->addAction(actLeft);
	menu->addAction(actRight);
	if (bFTS == true) {							// showing fit to screen
		menu->addAction(actView0);
	} else {
		menu->addAction(actView1);
	}
	menu->addSeparator();
	menu->addAction(actSaveImg);
	menu->addSeparator();
	menu->addAction(actDlg);
	menu->addSeparator();
	menu->addAction(actHelp);
	menu->addAction(actAbout);
	menu->addSeparator();
	menu->addAction(actQuit);
}

void Siv::menuToggleFitToScreen()			// show original size, only proceed if image is larger than screen in current rotated view
{												// check if original size > screen dimension, if not, nothing to do
	if (iRotate == 0 or iRotate == 2) {			// image shown horizontally
		if(iImOrigX < iScrX + 1 and iImOrigY < iScrY + 1) {
			return;
		}
	} else if (iRotate == 1 or iRotate == 3) {	// image shown vertically (+/- 90° rotated)
		if (iImOrigY < iScrX + 1 and iImOrigX < iScrY + 1) {
			return;
		}
	}

	if (bFTS == true) {							// showing FitToScreen
		bFTS = false;
	} else {									// showing original size
		bFTS = true;
	}
	usrImageChange();
}

void Siv::menuLeft()						// rotate image 90° to left
{
	iRotate = iRotate - 1;
	if (iRotate < 0) {
		iRotate = 3;
	}
	usrImageChange();							// reload current image
}

void Siv::menuRight()						// rotate image 90° to right
{
	iRotate = iRotate + 1;
	if (iRotate > 3) {
		iRotate = 0;
	}
	usrImageChange();							// reload current image
}


// ********************************************************************************************************************
//  **********  message boxes  **********
// **************************************
int Siv::messageBox(int iD, const QString& sD, const QString& sD2)	// show QMessageBox at cursor position, type 1=info, 2= warn, 3=critical, 4=question
{
	QMessageBox msgBox;
	msgBox.setWindowTitle(sD);
	if (iD == 1) {
		msgBox.setIcon(QMessageBox::Information);
	} else if (iD == 2) {
		msgBox.setIcon(QMessageBox::Warning);
	} else if (iD == 3) {
		msgBox.setIcon(QMessageBox::Critical);
	} else if (iD == 4) {
		msgBox.setIcon(QMessageBox::Question);
	} else {
		return -1;
	}
	msgBox.setText(sD2);
	if (iD < 4) {
		msgBox.addButton(tr("&OK"), QMessageBox::AcceptRole);
	} else {
		msgBox.addButton(QMessageBox::Yes);
		msgBox.addButton(QMessageBox::No);
		msgBox.setDefaultButton(QMessageBox::No);
	}
	msgBox.move(getCursorPos());
	return msgBox.exec();
}

QPoint Siv::getCursorPos()					// get cursor position - iX=x , iY=y
{
	qPos = QCursor::pos();
	qPos.setX(qPos.x() - 190);
	qPos.setY(qPos.y() - 125);
	return qPos;
}


// *****************************************
// **  user selected image functions      **
// *****************************************
void Siv::usrImageNew(const QString& sD)	// show new image           *****************************need to set wait cursor here or leave, just waste of time?
{
	if (timerAnimNext->isActive() == true) {
		timerAnimNext->stop();
	}
	setCurType(1);								// wait cursor
	imLoad(sD);
//	imRotateGetExifHint();
//	imRotate();
	imResize();
	imShow();
	showImNameSize();
	setCurGrab();
	setCurType(0);								// arrow cursor
	imInfoUpd();
}

void Siv::usrImageChange()					// resize (FTS/orig size) or change rotation of current image on user request
{                                                                       // *****************************need to set wait cursor here or leave, just waste of time?
	setCurType(1);
	imLoad(sImCur);
//	imRotate();
	imResize();
	imShow();
	showImNameSize();
	setCurGrab();
	setCurType(0);
	imInfoToFront();
	menuPop();
}


// *****************************************
// **  image functions                    **
// *****************************************
void Siv::imLoad(const QString& sD)         // try loading image sD, full path and name ************************simplify, improve code here
{
	sImCur = sD;
	bAnimated = false;
	imReader.setFileName(sImCur);
		imReader.setAutoTransform(true);		// rotate according EXIF info (if any)

	qiCur = imReader.read();
	if (qiCur.isNull()) {						// error loading image
		bLoadNext = true;
		imLoadErr();
		return;
	}

	if (imReader.format() == "gif" and imReader.imageCount() > 0) {	// check if animated gif, if yes start loop through images
		bAnimated = true;
		iAnimatedImgCur = 0;
		iAnimatedImgMax = imReader.imageCount();
		iAnimatedDelay = imReader.nextImageDelay();
		if (timerAnimNext->isActive() == true) {
			timerAnimNext->stop();
		}
		timerAnimNext->start(iAnimatedDelay);
	}
		//    cout << "imLoad: qiCur: width, height, depth: " << QString::number( qiCur.width() ).toStdString() << ":" << QString::number( qiCur.height() ).toStdString() << ":" << QString::number( qiCur.depth() ).toStdString() << endl;                                     // *****  debug  *****
	qiCurC = qiCur;
	bLoadNext = false;
	iImOrigX = qiCur.width();				// width of original image before any manipulation
	iImOrigY = qiCur.height();				// heigth of original image before any manipulation
	iImCurX = iImOrigX;
	iImCurY = iImOrigY;
	iImLastGood = iImCurListIdx;			// remember last file we could load
}

void Siv::imLoadErr()						// error while trying to load image file, ask to try next or abort
{
//	iWidth = 550;
//    iHeight = 240;
	// show QMessageBox: type 1=info, 2=warn, 3=critical, 4=question - at current cursor position - on question returns 5=yes or 6=no
	iD = messageBox(4, tr("Error - try next image?"), tr("<html><body>The file<br>"
														"%1<br>"
														"could not be loaded!<p>"
														"It will be deleted from the image list.<p>"
														"Do you want to try the next image in the list (if any)?</body></html>").arg(sImCur));
	if (iD == 6) {
		bLoadNext == false;
		removeImCurFromList();					// now iImCurListIdx already pointing to next image !
		if (iImCurListIdx < 0) {
			iImCurListIdx = 0;
		}
		return;
	}

	bLoadNext == true;
	if (bInit == true) {						// just started, loading first image
		iImLastGood = -1;						// none so far
		init();
	}
	removeImCurFromList();						// now iImCurListIdx already pointing to next image !
	requestedNewImage(1);
}

void Siv::removeImCurFromList()				// remove current image from image list, failed to load
{
	slFiles.removeAt(iImCurListIdx);
	if (iImLastGood > iImCurListIdx) {			// moving backwards in list
		iImLastGood = iImLastGood - 1;
	} else {
		iImCurListIdx = iImCurListIdx - 1;
		if (iImCurListIdx < 0) {				// removed first image
			iImCurListIdx = 0;
		}
	}
	if (slFiles.count() == 0) {					// fatal, no images left in list, could not load any
		bQuit = true;
		sMBTit = tr("siimvi - fatal error nothing to show !");
		sMBTxt = tr("The file\n%1\ncould not be loaded!\nThere is no more file to try. Aborting...\n \n").arg(sImCur);
		wCtrl(3);
	}
	sImCur = slFiles.at(iImCurListIdx);
	showImNameSize();
}

void Siv::imRotateGetExifHint()				// check Exif info if need to rotate, not set here, set in next step
{
	iRotate = 0;
	if (bExif == true) {						// at first image load
		try{
			Exiv2::Image::UniquePtr imageE = Exiv2::ImageFactory::open(sImCur.toStdString());	// start reading exif information with exiv2
			assert (imageE.get() !=0);
			imageE->readMetadata();
			Exiv2::ExifData &exifData = imageE->exifData();
			if (exifData.empty() != true) {													// only try setting exif info, if any in image
				Exiv2::ExifKey key("Exif.Image.Orientation");									// orientation
				Exiv2::ExifData::iterator pos = exifData.findKey(key);
				if (pos != exifData.end()) {
					sD = QString::fromStdString(pos->value().toString());
					if (sD == "3" or sD == "4") {												// rotate 180°
						iRotate = 2;
					} else if (sD == "5" or sD == "6") {										// rotate 90° right
						iRotate = 1;
					} else if (sD == "7" or sD == "8") {										// rotate 90° left
						iRotate = 3;
					}
				}
			}
		}
		catch (Exiv2::Error& e) {
//cout << "imRotateGetExifHint(): caught error" << endl;                                               // *****  debug  *****
			iRotate = 99;                      //      iExiv2 = 0;
		}
	}
}

void Siv::imRotate()						// rotate image according exif corrections and user actions
{
	if (iRotate > 0 and iRotate < 99) {
		setCurType(1);
//		QMatrix4x4 matrix;
		QTransform transform;
		switch (iRotate) {
			case 1:								// rotate 90° right
//				matrix = matrix.rotate(90, QVector3D(0, 0, 1));
				transform.rotate(90);
				qiCur = qiCur.transformed(transform, Qt::SmoothTransformation);	// Qt::FastTransformation
				iImCurX = qiCur.width();		// width of current image after rotation
				iImCurY = qiCur.height();		// heigth of current image after rotation
				break;
			case 2:
				transform.rotate(180);
				qiCur = qiCur.transformed(transform, Qt::SmoothTransformation);	// Qt::FastTransformation
				iImCurX = qiCur.width();		// width of current image after rotation
				iImCurY = qiCur.height();		// heigth of current image after rotation
				break;
			case 3:
				transform.rotate(-90);
				qiCur = qiCur.transformed(transform, Qt::SmoothTransformation);	// Qt::FastTransformation
				iImCurX = qiCur.width();		// width of current image after rotation
				iImCurY = qiCur.height();		// heigth of current image after rotation
				break;
		}
	}
}

void Siv::imResize()						// scale/resize image to fit to screen or original size view
{
	if (bFTS == true and bAnimated == false) {		// do not resize animated images
		if (bInit == false) {
			timerHideCurStop();						// stop 2 sec timer, no longer in grab mode
		}
		if (iImCurX > iScrX or iImCurY > iScrY) {	// only scale if image > screen
			dX = double(iScrX) / double(iImCurX);
			dY = double(iScrY) / double(iImCurY);
			dD = min( dX, dY);
			iImCurX = iImCurX * dD;
			iImCurY = iImCurY * dD;
			qiCur = qiCur.scaled(iImCurX, iImCurY, Qt::IgnoreAspectRatio, Qt::SmoothTransformation);	// Qt::FastTransformation
			sD = QString::number(iImCurX) + "x" + QString::number(iImCurY);
		}
	}
}

void Siv::imShow()							// show image
{
	qpmCur = QPixmap::fromImage(qiCur);
	iImShownX = qpmCur.width();					// get image size imGetSize
	iImShownY = qpmCur.height();

	iScrStartX = iScrStartY = 0;				// re-set position to start drawing image on screen to 0
	iImStartX = iImStartY = 0;					// re-set position within the image to draw to 0

	imAlign();

	setUpdatesEnabled(true);
	update();									// triggers qpaintevent
	setCurType(0);								// arrow cursor
	timerHideCurStart();
//  bBlockWheel = false;                          // allow next image to be selected
//cout << "end of image show" << endl;                                               // *****  debug  *****
}

void Siv::imShowAnimNext()					// show next image of an animated image
{
	iAnimatedImgCur = iAnimatedImgCur + 1;
	if (iAnimatedImgCur > iAnimatedImgMax - 2) {	// image 0 to max -2 = we don't show the last image - reduces flicker
		timerAnimNext->stop();
		imLoad(sImCur);
	}

//cout << "imShowAnimNext - iAnimatedImgCur = " << QString::number(iAnimatedImgCur).toStdString() << " , iAnimatedImgMax = " << QString::number(iAnimatedImgMax).toStdString() << endl;                                               // *****  debug  *****
//cout << "currentImageNumber = " << QString::number(imReader.currentImageNumber()).toStdString() << endl;                                               // *****  debug  *****
	qiCur = imReader.read();
	imShow();
}

void Siv::imAlign()							// adjust image to center of screen or set crop to show (if image larger then screen)
{
	if (iImShownX <= iScrX) {							// image < screen: center in x
		iScrStartX = int((iScrX - iImShownX) / 2);		// top left x coordinate
		bGrabX = false;
	} else {											// image > screen: show upper left
		bGrabX = true;
	}

	if (iImShownY <= iScrY) {							// image < screen: center in y
		iScrStartY = int((iScrY - iImShownY) / 2);		// top left y coordinate
		bGrabY = false;
	} else {											// image > screen: show upper left
		bGrabY = true;
	}
}

void Siv::paintEvent(QPaintEvent *)
{
	QPainter painter(this);
	painter.drawPixmap(iScrStartX, iScrStartY, qpmCur, iImStartX, iImStartY, iImShownX, iImShownY);	// iPdTopLeftX,Y , pixmap, iPmTopLeftX,Y , iPmSizeX,Y
}

void Siv::showImNameSize()					// show name and/or size of image a short time
{
	if (bShowName == true or bShowSize == true) {		// show image name and/or size
		sTt = " " + QString::number(iImCurListIdx + 1) + "/" + QString::number(slFiles.count());
		if (bShowName == true) {
			sTt = sTt + ", " + sImCur;
		}
		if (bShowSize == true) {
			sTt = sTt + ", " + QString::number(iImOrigX) + "x" + QString::number(iImOrigY) + " at " + QString::number(iImCurX) + "x" + QString::number(iImCurY);
		}
		sTt = sTt + " ";
		emit sigShowNameSize(iScrX, iScrY, sTt);
	}
}


// *****************************************
// **  cursor handling                    **
// *****************************************
void Siv::setCurGrab()                      // set cursor grab mode
{
  if (bGrabX == false) {                           // image < screen: center in x
    iMoX = iMLimLowX = iMLimUpX = int(iScrX / 2);     // set mouse x to mid of screen
  } else {                                          // image > screen: show upper left
    iScrX3 = int(iScrX / 3);
    iMaxOffX = int(iImShownX - iScrX);                   // length that image is wider than screen in X
    if (iMaxOffX > iScrX3) {                           // image larger screen plus 1/3 of screen: need move factor > 1
      iMFactX = iMaxOffX / int(iScrX / 3) + 1;
    } else {
      iMFactX = 1;
    }
    iMLimLowX = int(iScrX / 2) - (iMaxOffX / 2 / iMFactX);
    iMLimUpX = int(iScrX / 2) + (iMaxOffX / 2 / iMFactX);
    iMoX = iMLimUpX;
  }

  if (bGrabY == false) {                           // image < screen: center in y
    iMoY = iMLimLowY = iMLimUpY = int(iScrY / 2);          // set mouse x to mid of screenint( uiScrY / 2 );
  } else {                                                // image > screen: show upper left
    iScrY3 = int(iScrY / 3);
    iMaxOffY = int(iImShownY - iScrY);
    if (iMaxOffY > iScrY3) {                             // image larger screen plus 1/3 of screen: need move factor > 1
      iMFactY = iMaxOffY / int(iScrY / 3) + 1;
    } else {
      iMFactY = 1;
    }
    iMLimLowY = int(iScrY / 2) - (iMaxOffY / 2 / iMFactY);
    iMLimUpY = int(iScrY / 2) + (iMaxOffY / 2 / iMFactY);
    iMoY = iMLimUpY;
  }

  if (bGrabX == true or bGrabY == true) {      // image larger than screen
    setCurType(2);                                    // closed hand cursor - grabbed
    if (bGrab == false) {
      bGrab = true;
      iCPX = iMoX;
      iCPY = iMoY;
      QTimer::singleShot(200, this, SLOT(setCurPos()));    // only in grab mode
      sTt = tr(" Image grabbed ");
      ttShow();
    }
  } else {                                                                            // image fits to screen
    setCurType(0);                                // arrow cursor
    if (bGrab == true) {
      bGrab = false;
      sTt = tr(" Image released ");
      ttShow();
    }
  }

  timerHideCurStart();
//  cout << "setCurGrab(): end reached" << endl;                                               // *****  debug  *****
//  bBlockWheel = false;                          // allow next image to be selected
}

void Siv::setCurPos()						// set cursor position
{
	QCursor().setPos(iCPX , iCPY);
}

void Siv::setCurType(int iC)				// set cursor type (image), 0=arrow, 1=wait, 2=closed hand, 3=blank
{
	if (iC < 3) {
		iLastVisCur = iC;
	}
	switch (iC) {
		case 0:									// arrow cursor
			setCursor(Qt::ArrowCursor);
			break;
		case 1:									// wait cursor
			setCursor(Qt::WaitCursor);
			break;
		case 2:									// closed hand
			setCursor(Qt::ClosedHandCursor);
			break;
		case 3:									// hide cursor
			setCursor(Qt::BlankCursor);
			break;
	}
}


// *****************************************
// **  event handling                     **
// *****************************************
void Siv::keyPressEvent(QKeyEvent *ke)      // key press event
{
	switch (ke->key()) {
		case Qt::Key_A:							// show about dialog
		    about();
			break;
		case Qt::Key_F:							// toggle show image in original size and fit to screen
//			menuFTS();
			break;
		case Qt::Key_H:							// show online help
		    help();
			break;
		case Qt::Key_I:							// show image information including soe Exif tag information
//  cout << "key i hit" << endl;                                               // *****  debug  *****
//			if (iWShown == 4) {
//			} else {
//				imInfoShow();
//			}
		    wCtrl(4);
			break;
		case Qt::Key_L:							// rotate image 90° left
			menuLeft();
			break;
		case Qt::Key_M:							// show menu
			menu->exec(QPoint(100, 100));
			break;
		case Qt::Key_Q:							// quit
			quitPrep();
			break;
		case Qt::Key_R:							// rotate image 90° right
			menuRight();
			break;
	    case Qt::Key_Down:						// next image
		    requestedNewImage(1);
		    break;
	    case Qt::Key_Escape:					// quit
		    quitPrep();
		    break;
	    case Qt::Key_PageDown:					// next image
		    requestedNewImage(1);
		    break;
	    case Qt::Key_PageUp:					// previous image
		    requestedNewImage(-1);
		    break;
	    case Qt::Key_Up:						// previous image
		    requestedNewImage(-1);
		    break;
	}
}

void Siv::mousePressEvent(QMouseEvent *me)	// mouse event
{
	switch (me->button()) {
		case Qt::LeftButton:						// toggle fit to screen
			if (iWShown == 0) {					// no widget shown
				menuToggleFitToScreen();
			}
			break;
		case Qt::MiddleButton:						// show image (exif) info
			if (bWInfo == false) {
				imInfoShow();
			} else {
				emit sigInfoHide();
			}
			break;
		case Qt::RightButton:						// show menu
			if (iWShown == 0) {
				menu->exec(QCursor::pos());
			}
			break;
	}
}

void Siv::wheelEvent(QWheelEvent *we)		// mouse wheel event
{
	if (bBlockWheel == false) {
		iD = we->angleDelta().y();
		if (iD > 15 or iD < -15) {				// broken, delta is in steps of 120, but wheel counts steps too fast, always only getting 1 step( +120/-120)
			bBlockWheel = true;					// block selecting next image until current is loaded or timed out (3 sec)
			if (iD > 0) {						// forward move, prev image
				requestedNewImage(-1);
			} else if (iD < 0) {				// backward move, next image
				requestedNewImage(1);
			}
		}
	}
}

void Siv::requestedNewImage(int iD)			// new image requested by mouse wheel, anything to show?
{
//  cout << "requestedNewImage(): reached " << QString::number( iD ).toStdString() << endl;                                //  ****** debug
	if (iD == 1) {								// next image available ?
		if (iImCurListIdx + 1 > slFiles.count() - 1) {
			if (slFiles.count() == 1) {
				sTt = tr(" No further image to show ");
			} else {
				sTt = tr(" At end of image list ");
			}
			ttShow();
			unBlockWheelTimer();
			return;
		}
	} else if (iD == -1) {						// prev image available ?
		if (iImCurListIdx - 1 < 0) {
			if (slFiles.count() == 1) {
				sTt = tr(" No further image to show ");
			} else {
				sTt = tr(" At begin of image list ");
			}
			ttShow();
			unBlockWheelTimer();
			return;
		}
	}
	unBlockWheelTimer();
	if (iWShown == 0) {
		iImCurListIdx = iImCurListIdx + iD;
		usrImageNew(slFiles.at(iImCurListIdx));
	}
//  cout << "requestedNewImage(): " << QString::number( iD ).toStdString() << " file to load: " << slFiles.at( iImCurListIdx ).toStdString() << endl;   // ***  debug  ***
}

void Siv::mouseMoveEvent(QMouseEvent *mme)	// mouse event
{
	if (bWSetDlg == false) {
		iMnX = mme->position().x();
		iMnY = mme->position().y();
		if (bGrabX == true) {
			if (iMnX >= iMLimLowX and iMnX <= iMLimUpX) {
				iMMoveX = iMnX - iMoX;						// mouse move in x
				iImMoveX = iMMoveX * iMFactX;				// image move in x (mouse move * factor)

				iImStartX = iImStartX - iImMoveX;			// define new image in x to show
				if (iImShownX - iImStartX < iScrX) {		// right border
					iImStartX = iImShownX - iScrX;
				} else if (iImStartX < 0) {
					iImStartX = 0;							// left border
				}
				iMoX = iMoX + iMMoveX;						// keep mouse cursor in x limits
			}
		}

		if (bGrabY == true) {
			if (iMnY >= iMLimLowY and iMnY <= iMLimUpY) {
				iMMoveY = iMnY - iMoY;						// mouse move in y
				iImMoveY = iMMoveY * iMFactY;				// image move in y (mouse move * factor)

				iImStartY = iImStartY - iImMoveY;			// define new image in y to show
				if (iImShownY - iImStartY < iScrY) {		// lower border
					iImStartY = iImShownY - iScrY;
				} else if (iImStartY < 0) {
					iImStartY = 0;							// upper border
				}
				iMoY = iMoY + iMMoveY;						// keep mouse cursor in x limits
			}
		}

		if (bGrab == true) {
			QCursor().setPos(iMoX , iMoY);
		}
		timerHideCurStart();
		update();
	}
}


// *****************************************
// **  timer handling                     **
// *****************************************
void Siv::unBlockWheelTimer()				// start 2,7 sec single shot timer to unblock the mouse wheel in case loading another image failed
{
	QTimer::singleShot(400, this, SLOT(unBlockWheel()));
}

void Siv::unBlockWheel()					// unblock the mouse wheel to allow moving to next image in case loading the other image takes more than 3 sec (failed?)
{
	bBlockWheel = false;
}

void Siv::timerAnimNextCreate()				// create timer to show next image of an animated image
{
	timerAnimNext = new QTimer(this);
	connect(timerAnimNext, SIGNAL(timeout()), this, SLOT(imShowAnimNext()));
}

void Siv::timerHideCurCreate()				// create 2 sec timer to hide mouse cursor in grab mode after last move
{
	timerHideCur = new QTimer(this);
	timerHideCur->setSingleShot(true);
	connect(timerHideCur, SIGNAL(timeout()), this, SLOT(timerHideCurTimedOut()));
}

void Siv::timerHideCurStart()
{
	if (timerHideCur->isActive() == true) {
		timerHideCur->stop();
	}
	setCurType(iLastVisCur);
	timerHideCur->start(800);
}

void Siv::timerHideCurStop()
{
	timerHideCur->stop();
}

void Siv::timerHideCurTimedOut()
{
	setCurType(3);                                  // hide cursor
}


// ********************************************************************************************************************
// **  tooltip handling             **
// ***********************************
void Siv::ttRec(const QString& sD)          // receive tooltip from other program parts ******************** used ????????????????????????????????
{
	sTt = sD;
	ttShow();
}

void Siv::ttShow()							// show tooltip that at begin or end of list
{
	qpD.setX(iMoX + 5);
	qpD.setY(iMoY - 10);
	emit sigShowListBeginEnd(qpD, sTt);
}


// ********************************************************************************************************************
// **  popup widget handling        **
// ***********************************
void Siv::wCtrl(int iWToShow)	// show popup widget, if no other is shown: 0=none, 1=about-abouthelp, 2=help-abouthelp, 3=error-abouthelp, 4=info, 5=settings dialog
{
//	if (iWShown == 0) {						// only proceed, if no other widget is shown
		switch(iWToShow) {
			case 1:									// show about widget
			    if (iWShown == 0) {
					iWShown = 1;
					ahwCreate();
					emit sigDlgAbout();
					break;
				} else if (iWShown == 1) {				// close widget
					ahwHidden();
					break;
				} else {
					break;								// other widget already shown - ignore request
				}
			case 2:									// show help widget
			    if (iWShown == 0) {
					iWShown = 2;
					ahwCreate();
					emit sigDlgHelp();
					break;
				} else if (iWShown == 2) {				// close widget
					ahwHidden();
					break;
				} else {
					break;								// other widget already shown - ignore request
				}
			case 3:									// show message widget
			    if (iWShown == 0) {
					iWShown = 3;
					ahwCreate();
					emit sigDlgMsg(sMBTit, sMBTxt);
					break;
				} else if (iWShown == 3) {				// close widget
					ahwHidden();
					break;
				} else {
					break;								// other widget already shown - ignore request
				}
			case 4:									// show info (exif) widget
			    if (iWShown == 0) {
					iWShown = 4;
					imInfoShow();
					break;
				} else if (iWShown == 4) {				// close widget
					imInfoHidden();
					break;
				} else {
					break;								// other widget already shown - ignore request
				}
			case 5:									// show settings dialog
			    if (iWShown == 0) {
					iWShown = 5;
					setCurType(0);
					setUpdatesEnabled(false);
					qpD = QCursor::pos();
					emit sigSetShow();
					break;
				} else if (iWShown == 5) {				// close widget
					setDlgClosed(true);					// play it save, re-read settings
					break;
				} else {
					break;								// other widget already shown - ignore request
				}
		}
//	}
}


// ********************************************************************************************************************
// **  about/help/error widget handling  **
// ****************************************
void Siv::about()							// about clicked
{
	wCtrl(1);
}

void Siv::help()							// help clicked
{
	wCtrl(2);
}

void Siv::ahwCreate()						// create about/help/error dialog widget
{
	if (bWAhw == false) {
		bWAhw = true;
		ahw = new aboutHelp();
		connect(this, SIGNAL(sigDlgAbout()), ahw, SLOT(showAbout()));
		connect(this, SIGNAL(sigDlgHelp()), ahw, SLOT(showHelp()));
		connect(this, SIGNAL(sigDlgMsg(const QString& , const QString&)), ahw, SLOT(showMsg(const QString& , const QString&)));
		connect(ahw, SIGNAL(sigAhwHidden()), this, SLOT(ahwHidden()));
	}
}

void Siv::ahwHidden()						// ahw widget was hidden
{
	ahw->deleteLater();
	bWAhw = false;
	iWShown = 0;
	if (bQuit) {
		quitPrep();
	}
}


// ********************************************************************************************************************
// **  info widget handling         **
// ***********************************
void Siv::imInfoShow()						// show/hide image information in popup window
{
	if (bWInfo == false) {						// create widget when requested
		bWInfo = true;
		inf = new InfoW();
		connect(this, SIGNAL(sigInfGet(const QString&, int, int, int, int, int)), inf, SLOT(iwUpd(const QString&, int, int, int, int, int)));
									//const QString& sImCur, int iImOrigX, int iImOrigY, int iImCurX, int iImCurY, int iRotate, int iDepth
		connect(this, SIGNAL(sigInfoToFrt()), inf, SLOT(dlgToFrt()));					// bring info dlg back to frt
		connect(this, SIGNAL(sigInfoHide()), inf, SLOT(dlgClose()));					// hide image info widget, new image selected
		connect(inf, SIGNAL(sigInfErr(const QString&, const QString&)), this, SLOT(imInfoErr(const QString&, const QString&)));
		connect(inf, SIGNAL(sigInfHidden()), this, SLOT(imInfoHidden()));				// image info widget hidden
		emit sigInfGet(sImCur, iImOrigX, iImOrigY, iImCurX, iImCurY, iRotate);
		menuPop();
	}
/* else {
    imInfoClose();
  }*/
}

void Siv::imInfoToFront()					// bring info widget back to front
{
	emit sigInfoToFrt();
}

void Siv::imInfoUpd()						// update image information
{
	if (bWInfo == true) {
		emit sigInfGet(sImCur, iImOrigX, iImOrigY, iImCurX, iImCurY, iRotate);
	}
}

void Siv::imInfoClose()						// close image information popup window
{
	emit sigInfoHide();
}

void Siv::imInfoHidden()					// image info widget was hidden
{
	inf->deleteLater();
	bWInfo = false;
	iWShown = 0;
	menuPop();
}

void Siv::imInfoErr(const QString& sD, const QString& sD2)	// error while getting exif tags for info widget - aborting
{
//  cout << "imInfoErr(): reached !" << sD2.toStdString() << endl;                                               // *****  debug  *****
	sMBTit = sD;
	sMBTxt = sD2;
	wCtrl(3);								// show error in message widget
}


// ********************************************************************************************************************
// **  settings dialog widget handling  **
// ***************************************
void Siv::setDlgShow()						// show settings dialog
{
	if (bWSetDlg == false) {
		bWSetDlg = true;
		setd = new setDlg();														// settings dialog
		connect(this, SIGNAL(sigSetShow()), setd, SLOT(setRead()));					// show settings dialog
		connect(setd, SIGNAL(sigReReadSet(bool)), this, SLOT(setDlgClosed(bool)));	// settings dialog closed, re-read settings, might have been changed
	}
	wCtrl(5);
}

void Siv::setDlgClosed(bool bD)			// settings dialog was closed
{
	QCursor::setPos(qpD);
	setUpdatesEnabled(true);
	setd->deleteLater();
	if (bD == true) {
		setRead();
		emit sigSetReadTt();
		emit sigSetReadTtNS();
		menuPop();
	}
	iWShown = 0;
	bWSetDlg = false;
}


// ********************************************************************************************************************
// **  save image                       **
// ***************************************
void Siv::saveImg()							// save current image
{
	saveImgSetNameType();
	saveImgGetFileName();
}

void Siv::saveImgSetNameType()				// save current image - set image name and type for file to save
{
	QFileInfo fiNT(sImCur);
		sFileBaseName = fiNT.completeBaseName();
		sFileSuffix = fiNT.suffix();
	if (sFileSuffix == "") {
		sFileSuffix = "jpg";
	}

// bmp and tif support missing - possible?
	if (sFileSuffix.indexOf("jpg", 0, Qt::CaseInsensitive)) {
		sImgType = "JPG";													// jpg to jpg
	} else if (sFileSuffix.indexOf("jpeg", 0, Qt::CaseInsensitive)) {
		sImgType = "JPEG";													// jpg to jpg
	} else if (sFileSuffix.indexOf("png", 0, Qt::CaseInsensitive)) {
		sImgType = "PNG";													// png to png
	} else if (sFileSuffix.indexOf("svg", 0, Qt::CaseInsensitive)) {
		sImgType = "SVG";													// svg to svg
	} else if (sFileSuffix.indexOf("tiff", 0, Qt::CaseInsensitive)) {
		sImgType = "TIFF";													// tiff to tiff - tiff not supported by QPixmap, Qt 4.5 ?
	} else {
		sImgType = "JPG";													// if not jpg, png or tiff, make it jpg
		sFileSuffix = "jpg";
	}
	sFileName = sFileBaseName + "." + sFileSuffix;
}

void Siv::saveImgGetFileName()				// save current image - get file name and path
{
	setStyleSheet("background-color: lightGray");
	QFileDialog fdlg(this);							// set file to save
		fdlg.setAcceptMode(QFileDialog::AcceptSave);
		fdlg.setFileMode(QFileDialog::AnyFile);
		fdlg.setNameFilter(tr("Images (*.jpg *jpeg *.png *.svg *.tiff *.xpm)"));
		fdlg.setViewMode(QFileDialog::Detail);
	QFileInfo finfo(sDir);							// sDir last directory used
	if (finfo.isReadable() != true) {
		sDir = QDir::homePath();					// sDir is the users home dir
	}
	fdlg.setDirectory(sDir);
	fdlg.selectFile(sFileName);

	if (fdlg.exec()) {
		slD = fdlg.selectedFiles();					// selected files
		sFileSaveAs = slD.at(0);					// 1st is the one we need
		saveImgSave();
	}
	setStyleSheet("background-color: black");
}

void Siv::saveImgSave()						// save current image - save do
{
	if (sFileSaveAs > "") {
		qiCurC.save(sFileSaveAs, sImgType.toUtf8() , 98);	// quality 98 seems to create image of same size as original for jpegs
		QFileInfo fiP(sFileSaveAs);
		sDir = fiP.absolutePath();
	}
}

